@@ -4,7 +4,7 @@ Generic Proxy Pattern Pool for Python.
 This code is public domain.
 """
 
-from typing import Optional, Callable, Dict, Set, Any
+from typing import Callable, Any
 from enum import Enum
 from dataclasses import dataclass
 from contextlib import contextmanager
@@ -14,23 +14,32 @@ import time
 import logging
 
 # get module version
-import pkg_resources as pkg  # type: ignore
+from importlib.metadata import version as pkg_version
 
-__version__ = pkg.require("ProxyPatternPool")[0].version
+__version__ = pkg_version("ProxyPatternPool")
 
 
 log = logging.getLogger("ppp")
 
 
-class PoolException(Exception):
+class PPPException(Exception):
+    """Common class for ProxyPatternPool exceptions."""
     pass
 
 
+class PoolException(PPPException):
+    """Common class for Pool exceptions."""
+    pass
+
+
+# NOTE this does not include the time to create a resource
 class TimeOut(PoolException):
+    """Timeout while acquiring a resource at the pool level."""
     pass
 
 
-class ProxyException(Exception):
+class ProxyException(PPPException):
+    """Common class for Proxy exceptions."""
     pass
 
 
@@ -43,8 +52,11 @@ class Pool:
     - timeout: give-up waiting after this time, None for no timeout.
     - max_use: how many times to use a something, 0 for unlimited.
     - max_avail_delay: remove objects if unused for this secs, 0.0 for unlimited.
-    - max_using_delay: warn if object is keept for more than this time, 0.0 for no warning.
+    - max_using_delay: warn if object is kept for more than this time, 0.0 for no warning.
+    - max_using_delay_kill: kill if object is kept for more than this time, 0.0 for no killing.
     - close: name of "close" method to call, if any.
+    - log_level: set logging level for local logger.
+    - tracer: generate debug information.
     """
 
     @dataclass
@@ -58,46 +70,55 @@ class Pool:
         fun: Callable[[int], Any],
         max_size: int = 0,
         min_size: int = 1,
-        timeout: float = None,
+        timeout: float|None = None,
         max_use: int = 0,
         max_avail_delay: float = 0.0,
         max_using_delay: float = 0.0,
-        close: Optional[str] = None,
-        # temporary upward compatibility
-        max_delay: float = 0.0,
+        max_using_delay_kill: float = 0.0,
+        close: str|None = None,
+        max_delay: float = 0.0,  # temporary upward compatibility
+        log_level: int|None = None,
+        tracer: Callable[[Any], str]|None = None,
     ):
-        # data attributes
+        # debugging
+        if log_level is not None:
+            log.setLevel(log_level)
+        self._tracer = tracer
+        # objects
         self._fun = fun
-        self._nobjs = 0
-        self._nuses = 0
-        self._ncreated = 0
+        self._nobjs = 0  # current number of objects
+        self._nuses = 0  # currenly in use
+        self._ncreated = 0  # total created
         self._max_size = max_size
         self._min_size = min_size
         self._timeout = timeout
-        self._max_use = max_use
+        self._max_use = max_use  # when to recycle
         self._max_avail_delay = max_avail_delay or max_delay
-        self._max_using_delay = max_using_delay
+        self._max_using_delay_kill = max_using_delay_kill
+        self._max_using_delay_warn = max_using_delay or max_using_delay_kill
+        if self._max_using_delay_kill and self._max_using_delay_warn > self._max_using_delay_kill:
+            log.warning("inconsistent max_using_delay_warn > max_using_delay_kill")
         self._close = close
         # pool's content: available vs in use objects
-        self._avail: Set[Any] = set()
-        self._using: Set[Any] = set()
+        self._avail: set[Any] = set()
+        self._using: set[Any] = set()
         # keep track of usage count and last ops
-        self._uses: Dict[Any, Pool.UseInfo] = {}
+        self._uses: dict[Any, Pool.UseInfo] = {}
         # global pool re-entrant lock to manage attributes
         self._lock = threading.RLock()
-        self._sem: Optional[threading.Semaphore] = None
+        self._sem: threading.Semaphore|None = None
         if self._max_size:
             self._sem = threading.BoundedSemaphore(self._max_size)
         # create the minimum number of objects
         while self._nobjs < self._min_size:
             self._new()
         # start housekeeper thread if needed
-        self._housekeeper: Optional[threading.Thread] = None
-        if self._max_avail_delay or self._max_using_delay:
+        self._housekeeper: threading.Thread|None = None
+        if self._max_avail_delay or self._max_using_delay_warn:
             self._delay = self._max_avail_delay
             if not self._delay or \
-               self._max_using_delay and self._delay > self._max_using_delay:  # fmt: skip
-                self._delay = self._max_using_delay
+               self._max_using_delay_warn and self._delay > self._max_using_delay_warn:  # fmt: skip
+                self._delay = self._max_using_delay_warn
             self._delay /= 2.0
         else:
             self._delay = 0.0
@@ -107,7 +128,7 @@ class Pool:
 
     def __str__(self):
         o, u, a, i = self._nobjs, self._nuses, len(self._avail), len(self._using)
-        return f"o={o} u={u} a={a} i={i}"
+        return f"objs={o} uses={u} avail={a} using={i} sem={self._sem}"
 
     def _now(self) -> float:
         """Return now as a convenient float, in seconds."""
@@ -118,27 +139,35 @@ class Pool:
         log.info(f"housekeeper running every {self._delay}")
         while True:
             time.sleep(self._delay)
-            log.debug(str(self))
-            if self._nobjs <= self._min_size and not self._max_using_delay:
-                continue
             with self._lock:
+                if log.getEffectiveLevel() == logging.DEBUG:
+                    log.debug(str(self))
+                    if self._tracer:
+                        for obj in self._avail:
+                            log.debug(f"avail: {self._tracer(obj)}")
+                        for obj in self._using:
+                            log.debug(f"using: {self._tracer(obj)}")
                 now = self._now()
-                if self._max_using_delay:
-                    # warn about long running objects
-                    long_running, long_time = 0, 0.0
+                if self._max_using_delay_warn:
+                    # kill long running objects
+                    long_run, long_kill, long_time = 0, 0, 0.0
                     for obj in list(self._using):
-                        if now - self._uses[obj].last_get >= self._max_using_delay:
-                            long_running += 1
-                            long_time += now - self._uses[obj].last_get
-                    if long_running:
-                        log.warning(
-                            f"long running objects: {long_running} ({long_time / long_running})"
-                        )
-                if self._max_avail_delay:
-                    # close objects unused for too long
+                        running = now - self._uses[obj].last_get
+                        if running >= self._max_using_delay_warn:
+                            long_time += running
+                            long_run += 1
+                        if self._max_using_delay_kill and running >= self._max_using_delay_kill:
+                            self._del(obj)
+                            long_kill += 1
+                    if long_run or long_kill:
+                        delay = (long_time / long_run) if long_run else 0.0
+                        log.warning(f"long running objects: {long_run} ({delay} seconds, {long_kill} killed)")
+                if self._max_avail_delay and self._nobjs > self._min_size:
+                    # close spurious objects unused for too long
                     for obj in list(self._avail):
                         if now - self._uses[obj].last_ret >= self._max_avail_delay:
                             self._del(obj)
+                            # stop deleting objects if min size is reached
                             if self._nobjs <= self._min_size:
                                 break
 
@@ -157,6 +186,7 @@ class Pool:
         log.debug(f"creating new obj with {self._fun}")
         with self._lock:
             if self._max_size and self._nobjs >= self._max_size:  # pragma: no cover
+                # this should not be raised thanks to the semaphore
                 raise PoolException(f"pool max size {self._max_size} reached")
             obj = self._fun(self._ncreated)
             self._ncreated += 1
@@ -185,8 +215,10 @@ class Pool:
 
     def get(self, timeout=None):
         """Get a object from the pool, possibly creating one if needed."""
+        # FIXME why while?
         while True:
             if self._sem:
+                # the acquired token will be released at the end of ret()
                 if not self._sem.acquire(timeout=timeout if timeout else self._timeout):
                     raise TimeOut(f"timeout after {timeout}")
             with self._lock:
@@ -213,13 +245,14 @@ class Pool:
                 self._uses[obj].last_ret = self._now()
                 self._avail.add(obj)
             if self._sem:
+                # release token acquired in get()
                 self._sem.release()
 
     @contextmanager
     def obj(self, timeout=None):
         """Extract one object from the pool in a `with` scope."""
         try:
-            o = self.get()
+            o = self.get(timeout)
             yield o
         finally:
             self.ret(o)
@@ -260,22 +293,28 @@ class Proxy:
         SHARED = 1
         THREAD = 2
         VERSATILE = 3
+        WERKZEUG = 3
+        GEVENT = 4
+        EVENTLET = 5
 
     def __init__(
         self,
         obj: Any = None,
         set_name: str = "set",
-        fun: Callable[[int], Any] = None,
+        fun: Callable[[int], Any]|None = None,
         max_size: int = 0,
         min_size: int = 1,
         max_use: int = 0,
         max_avail_delay: float = 0.0,
         max_using_delay: float = 0.0,
-        timeout: float = None,
+        max_using_delay_kill: float = 0.0,
+        timeout: float|None = None,
         scope: Scope = Scope.AUTO,
-        close: Optional[str] = None,
+        close: str|None = None,
         # temporary backward compatibility
         max_delay: float = 0.0,
+        log_level: int|None = None,
+        tracer: Callable[[Any], str]|None = None,
     ):
         """Constructor parameters:
 
@@ -287,11 +326,16 @@ class Proxy:
         - max_use: when pooling, how many times to reuse an object.
         - max_avail_delay: when pooling, when to discard an unused object.
         - max_using_delay: when pooling, warn about long uses.
+        - max_using_delay_kill: when pooling, kill long uses.
         - timeout: when pooling, how long to wait for an object.
         - scope: level of sharing, default is to chose between SHARED and THREAD.
         - close: "close" method, if any.
+        - log_level: set logging level for local logger.
+        - tracer: generate debug information.
         """
         # scope encodes the expected object unicity or multiplicity
+        if log_level is not None:
+            log.setLevel(log_level)
         self._scope = (
             Proxy.Scope.SHARED if scope == Proxy.Scope.AUTO and obj else
             Proxy.Scope.THREAD if scope == Proxy.Scope.AUTO and fun else
@@ -301,8 +345,10 @@ class Proxy:
         self._pool_max_use = max_use
         self._pool_max_avail_delay = max_avail_delay or max_delay
         self._pool_max_using_delay = max_using_delay
+        self._pool_max_using_delay_kill = max_using_delay_kill
         self._pool_timeout = timeout
-        self._close = close
+        self._pool_tracer = tracer
+        self._pool_close = close
         self._set(obj=obj, fun=fun, mandatory=False)
         if set_name and set_name != "_set":
             setattr(self, set_name, self._set)
@@ -324,7 +370,8 @@ class Proxy:
         """Set current wrapped object generation function."""
         if self._scope == Proxy.Scope.AUTO:
             self._scope = Proxy.Scope.THREAD
-        assert self._scope in (Proxy.Scope.THREAD, Proxy.Scope.VERSATILE)
+        assert self._scope in (Proxy.Scope.THREAD, Proxy.Scope.VERSATILE,
+            Proxy.Scope.WERKZEUG, Proxy.Scope.EVENTLET, Proxy.Scope.GEVENT)
         self._fun = fun
         self._pool = Pool(fun,
                           max_size=self._pool_max_size,
@@ -333,21 +380,36 @@ class Proxy:
                           max_use=self._pool_max_use,
                           max_avail_delay=self._pool_max_avail_delay,
                           max_using_delay=self._pool_max_using_delay,
-                          close=self._close) \
+                          max_using_delay_kill=self._pool_max_using_delay_kill,
+                          close=self._pool_close,
+                          tracer=self._pool_tracer) \
             if self._pool_max_size is not None else None  # fmt: skip
         self._nobjs = 0
+
+        # local implementation (*event coverage skip for 3.12)
         if self._scope == Proxy.Scope.THREAD:
             self._local = threading.local()
-        else:  # Proxy.Scope.VERSATILE
+        elif self._scope == Proxy.Scope.WERKZEUG:
             from werkzeug.local import Local
 
             self._local = Local()
+        elif self._scope == Proxy.Scope.GEVENT:  # pragma: no cover
+            from gevent.local import local  # type: ignore
+
+            self._local = local()
+        elif self._scope == Proxy.Scope.EVENTLET:  # pragma: no cover
+            from eventlet.corolocal import local  # type: ignore
+
+            self._local = local()
+        else:  # pragma: no cover
+            raise Exception(f"unexpected local scope: {self._scope}")
+
         return self
 
     def _set(
         self,
         obj: Any = None,
-        fun: Optional[Callable[[int], Any]] = None,
+        fun: Callable[[int], Any]|None = None,
         mandatory=True,
     ):
         """Set current wrapped object or generation function."""
